﻿using System;
using System.Configuration;
using System.Web.Configuration;
using MvcErrorHandling.DomainModel.AppSettingsModel;
using MvcErrorHandling.DomainModel.CustomErrorsModel;
using MvcErrorHandling.DomainModel.HttpErrorsModel;
using MvcErrorHandling.DomainModel.HttpRuntimeModel;

namespace MvcErrorHandling.DomainModel
{
    public class ErrorConfiguration
    {
        private readonly Configuration _configuration;

        public CustomErrors CustomErrors { get; private set; }

        public HttpRuntimeSettings HttpRuntime { get; private set; }

        public HttpErrors HttpErrors { get; private set; }

        public AppSettingsWriter AppSettingsWriter { get; private set; }

        public ErrorConfiguration(string virtualPath)
        {
            _configuration = OpenWebConfiguration("~");

            var customErrorsSection = GetCustomErrorsSection();
            CustomErrors = new CustomErrors(customErrorsSection);

            var httpRuntimeSection = GetHttpRuntimeSection();
            HttpRuntime = new HttpRuntimeSettings(httpRuntimeSection);

            var systemWebServerSection = GetSystemWebServerSection();
            HttpErrors = new HttpErrors(systemWebServerSection, virtualPath);

            var appSettings = GetAppSettingsSection();
            AppSettingsWriter = new AppSettingsWriter(appSettings);
        }

        public void Read()
        {
            CustomErrors.ReadSection();
            HttpRuntime.ReadSection();
            HttpErrors.ReadSection();
            AppSettingsWriter.ReadSettings();
        }

        public void Reset()
        {
            CustomErrors.UpdateSection();
            HttpRuntime.UpdateSection();
            HttpErrors.UpdateSection();
            AppSettingsWriter.UpdateSettings();
        }

        public void Persist()
        {
            _configuration.Save(ConfigurationSaveMode.Minimal);
        }

        private static Configuration OpenWebConfiguration(string path)
        {
            var configuration = WebConfigurationManager.OpenWebConfiguration(path);

            if (configuration == null)
            {
                throw new Exception(string.Format("Unable to open Web configuration with path {0}", path));
            }

            return configuration;
        }

        private CustomErrorsSection GetCustomErrorsSection()
        {
            var systemWebGroup = GetSystemWebGroup();

            var customErrorsSection = systemWebGroup.Sections["customErrors"] as CustomErrorsSection;

            if (customErrorsSection == null)
            {
                throw new Exception("Unable to get customErrors section from Web configuration.");
            }

            return customErrorsSection;
        }

        private HttpRuntimeSection GetHttpRuntimeSection()
        {
            var systemWebGroup = GetSystemWebGroup();

            var httpRuntimeSection = systemWebGroup.Sections["httpRuntime"] as HttpRuntimeSection;

            if (httpRuntimeSection == null)
            {
                throw new Exception("Unable to get httpRuntime section from Web configuration.");
            }

            return httpRuntimeSection;
        }

        private ConfigurationSectionGroup GetSystemWebGroup()
        {
            var systemWebGroup = _configuration.GetSectionGroup("system.web");

            if (systemWebGroup == null)
            {
                throw new Exception("Unable to get system.web section group from Web configuration.");
            }

            return systemWebGroup;
        }

        private ConfigurationSection GetSystemWebServerSection()
        {
            var systemWebServer = _configuration.GetSection("system.webServer");

            return systemWebServer;
        }

        private KeyValueConfigurationCollection GetAppSettingsSection()
        {
            return _configuration.AppSettings.Settings;
        }
    }
}