﻿using System;
using System.Collections.Generic;
using System.Configuration;
using System.Linq;
using System.Xml.Linq;

namespace MvcErrorHandling.DomainModel.HttpErrorsModel
{
    public class HttpErrors
    {
        private readonly ConfigurationSection _systemWebServerSection;
        private readonly string _virtualPath;
        private HttpErrorsError _forbiddenError;
        private HttpErrorsError _notFoundError;
        private HttpErrorsError _serverError;

        public HttpErrorsMode Mode { get; private set; }
        public HttpErrorsExistingResponseMode ExistingResponseMode { get; private set; }
        public HttpErrorsRedirectType ForbiddenErrorRedirect { get; private set; }
        public HttpErrorsRedirectType NotFoundErrorRedirect { get; private set; }
        public HttpErrorsRedirectType ServerErrorRedirect { get; private set; }

        public HttpErrors(ConfigurationSection systemWebServerSection, string virtualPath)
        {
            if (systemWebServerSection == null) throw new ArgumentNullException("systemWebServerSection");
            if (string.IsNullOrWhiteSpace(virtualPath)) throw new ArgumentException("virtualPath");

            _systemWebServerSection = systemWebServerSection;
            _virtualPath = virtualPath;
        }

        public void ReadSection()
        {
            _forbiddenError = new HttpErrorsError(403, _virtualPath);
            _notFoundError = new HttpErrorsError(404, _virtualPath);
            _serverError = new HttpErrorsError(500, _virtualPath);

            ReadHttpErrorsSettings();
        }

        public void UpdateSection()
        {
            UpdateSection(
                EnumHelpers.GetDefaultEnumValue<HttpErrorsMode>(),
                EnumHelpers.GetDefaultEnumValue<HttpErrorsExistingResponseMode>(),
                EnumHelpers.GetDefaultEnumValue<HttpErrorsRedirectType>(),
                EnumHelpers.GetDefaultEnumValue<HttpErrorsRedirectType>(),
                EnumHelpers.GetDefaultEnumValue<HttpErrorsRedirectType>());
        }

        public void UpdateSection(
            HttpErrorsMode httpErrorsMode,
            HttpErrorsExistingResponseMode existingResponseMode,
            HttpErrorsRedirectType notFoundErrorRedirect,
            HttpErrorsRedirectType forbiddenErrorRedirect,
            HttpErrorsRedirectType serverErrorRedirect)
        {
            Mode = httpErrorsMode;
            ExistingResponseMode = existingResponseMode;
            NotFoundErrorRedirect = notFoundErrorRedirect;
            ForbiddenErrorRedirect = forbiddenErrorRedirect;
            ServerErrorRedirect = serverErrorRedirect;

            var httpErrorsElement = new XElement("httpErrors");

            httpErrorsElement.SetAttributeValue("errorMode", Mode.ToString());
            httpErrorsElement.SetAttributeValue("existingResponse", ExistingResponseMode.ToString());

            var childElements = new List<XElement>();

            childElements.AddRange(_forbiddenError == null ? new List<XElement>() : _forbiddenError.CreateErrorXmlElements(ForbiddenErrorRedirect));
            childElements.AddRange(_notFoundError == null ? new List<XElement>() : _notFoundError.CreateErrorXmlElements(NotFoundErrorRedirect));
            childElements.AddRange(_serverError == null ? new List<XElement>() : _serverError.CreateErrorXmlElements(ServerErrorRedirect));

            httpErrorsElement.Add(childElements);

            WriteHttpErrorsSettings(httpErrorsElement);
        }

        private void WriteHttpErrorsSettings(XElement httpErrorsElement)
        {
            var systemWebServerElement = GetSystemWebServerSectionElement();

            var httpErrors = systemWebServerElement.Descendants("httpErrors").FirstOrDefault();

            if (httpErrors != null)
            {
                httpErrors.Remove();
            }

            systemWebServerElement.Add(httpErrorsElement);

            var systemWebServerXml = systemWebServerElement.ToString();

            _systemWebServerSection.SectionInformation.SetRawXml(systemWebServerXml);
        }

        private XElement GetSystemWebServerSectionElement()
        {
            var systemWebServerRawXml = _systemWebServerSection.SectionInformation.GetRawXml();

            var systemWebServerXml = XElement.Parse(systemWebServerRawXml, LoadOptions.PreserveWhitespace);

            return systemWebServerXml;
        }

        private void ReadHttpErrorsSettings()
        {
            var systemWebServerElement = GetSystemWebServerSectionElement();

            var httpErrors = systemWebServerElement.Descendants("httpErrors").FirstOrDefault();

            if (httpErrors == null)
            {
                Mode = EnumHelpers.GetDefaultEnumValue<HttpErrorsMode>();
                ExistingResponseMode = EnumHelpers.GetDefaultEnumValue<HttpErrorsExistingResponseMode>();
                ForbiddenErrorRedirect = EnumHelpers.GetDefaultEnumValue<HttpErrorsRedirectType>();
                NotFoundErrorRedirect = EnumHelpers.GetDefaultEnumValue<HttpErrorsRedirectType>();
                ServerErrorRedirect = EnumHelpers.GetDefaultEnumValue<HttpErrorsRedirectType>();

                return;
            }

            var errorModeAttribute = httpErrors.Attribute("errorMode");
            Mode = ParseEnumValue<HttpErrorsMode>(errorModeAttribute);

            var existingResponseAttribute = httpErrors.Attribute("existingResponse");
            ExistingResponseMode = ParseEnumValue<HttpErrorsExistingResponseMode>(existingResponseAttribute);

            ForbiddenErrorRedirect = _forbiddenError.ParseHttpErrorsRedirectType(httpErrors);
            NotFoundErrorRedirect = _notFoundError.ParseHttpErrorsRedirectType(httpErrors);
            ServerErrorRedirect = _serverError.ParseHttpErrorsRedirectType(httpErrors);
        }

        private static TEnum ParseEnumValue<TEnum>(XAttribute attribute)
        {
            if (attribute == null)
            {
                return EnumHelpers.GetDefaultEnumValue<TEnum>();
            }

            return (TEnum)Enum.Parse(typeof(TEnum), attribute.Value, true);
        }
    }
}