﻿using System;
using System.Collections.Generic;
using System.Xml.Linq;
using System.Xml.XPath;

namespace MvcErrorHandling.DomainModel.HttpErrorsModel
{
    public class HttpErrorsError
    {
        private const string StaticPageFormatString = "{0}HttpErrors{1}.htm";
        private const string MvcPageFormatString = "{0}Error/HttpErrors{1}";
        private const string AspxPageFormatString = "{0}HttpErrors{1}.aspx";

        private readonly int _statusCode;
        private readonly string _htmPage;
        private readonly string _mvcPage;
        private readonly string _aspxPage;

        public HttpErrorsError(int statusCode, string virtualPath)
        {
            _statusCode = statusCode;
            _htmPage = string.Format(StaticPageFormatString, virtualPath, _statusCode);
            _mvcPage = string.Format(MvcPageFormatString, virtualPath, _statusCode);
            _aspxPage = string.Format(AspxPageFormatString, virtualPath, _statusCode);
        }

        public HttpErrorsRedirectType ParseHttpErrorsRedirectType(XElement httpErrors)
        {
            var errorElementXPath = string.Format("error[@statusCode='{0}']", _statusCode);

            var errorElement = httpErrors.XPathSelectElement(errorElementXPath);

            if (errorElement == null)
            {
                return EnumHelpers.GetDefaultEnumValue<HttpErrorsRedirectType>();
            }

            var pathAttribute = errorElement.Attribute("path");

            if (pathAttribute == null)
            {
                throw new Exception(string.Format("Failed to match redirect for status code {0}.", _statusCode));
            }

            var path = pathAttribute.Value;

            if (string.Equals(path, _htmPage, StringComparison.OrdinalIgnoreCase))
            {
                return HttpErrorsRedirectType.StaticErrorPage;
            }

            if (string.Equals(path, _mvcPage, StringComparison.OrdinalIgnoreCase))
            {
                return HttpErrorsRedirectType.MvcErrorPage;
            }

            if (string.Equals(path, _aspxPage, StringComparison.OrdinalIgnoreCase))
            {
                return HttpErrorsRedirectType.AspxErrorPage;
            }

            throw new Exception(string.Format("Failed to match redirect {0} for status code {1}.", path, _statusCode));
        }

        public List<XElement> CreateErrorXmlElements(HttpErrorsRedirectType httpErrorsRedirectType)
        {
            if (httpErrorsRedirectType == HttpErrorsRedirectType.NoPage)
            {
                return new List<XElement>();
            }

            string path;
            string responseMode;

            switch (httpErrorsRedirectType)
            {
                case HttpErrorsRedirectType.StaticErrorPage:
                    path = _htmPage;
                    responseMode = "File";
                    break;
                case HttpErrorsRedirectType.MvcErrorPage:
                    path = _mvcPage;
                    responseMode = "ExecuteURL";
                    break;
                case HttpErrorsRedirectType.AspxErrorPage:
                    path = _aspxPage;
                    responseMode = "ExecuteURL";
                    break;
                default:
                    throw new ArgumentOutOfRangeException("httpErrorsRedirectType");
            }

            var removeElement = new XElement("remove");
            removeElement.SetAttributeValue("statusCode", _statusCode);

            var errorElement = new XElement("error");
            errorElement.SetAttributeValue("statusCode", _statusCode);
            errorElement.SetAttributeValue("path", path);
            errorElement.SetAttributeValue("responseMode", responseMode);

            return new List<XElement> {removeElement, errorElement};
        }
    }
}