﻿using System;
using System.Configuration;
using System.Net;
using System.Web;
using System.Web.Mvc;
using System.Web.Routing;
using MvcErrorHandling.DomainModel.AppSettingsModel;

namespace MvcErrorHandling.MvcHelpers
{
    /// <summary>
    /// Rather than add the Application_Error handler to global.asax, I have created an HttpModule specifically for it.
    /// </summary>
    // ReSharper disable InconsistentNaming
    public class ErrorHandlerHttpModule : IHttpModule
    {
        private const string ErrorControllerRouteName = "Error";

        public void Init(HttpApplication context)
        {
            context.Error += Application_Error;
        }

        private static void Application_Error(object sender, EventArgs e)
        {
            var handleServerErrors =
                ConfigurationManager.AppSettings.Include(ManualServerErrorHandlingMode.ApplicationErrorHandler);

            var handleNotFoundErrors =
                ConfigurationManager.AppSettings.Include(ManualNotFoundErrorHandlingMode.ApplicationErrorHandler);

            if (!handleServerErrors && !handleNotFoundErrors)
            {
                return;
            }

            // Get the last error.

            var exception = HttpContext.Current.Server.GetLastError();

            if (exception is HttpUnhandledException)
            {
                exception = exception.InnerException ?? exception;
            }

            var httpException = exception as HttpException;

            // The exception can be logged here if you want.

            // Determine if the error should be handled by this module, and if so, how.

            var statusCode = httpException != null 
                ? httpException.GetHttpCode()
                : (int) HttpStatusCode.InternalServerError;

            if (statusCode == (int)HttpStatusCode.NotFound)
            {
                if (!handleNotFoundErrors)
                {
                    return;
                }
            }
            else if (!handleServerErrors)
            {
                return;
            }

            if (ConfigurationManager.AppSettings.Include(ErrorPageExceptionsMode.HandleInApplicationError))
            {
                // Handle the case where an error page threw an exception; we don't want to loop back to that code.

                if (HttpContext.Current.Request.Path.StartsWithIgnoreCase("/" + ErrorControllerRouteName + "/"))
                {
                    HandleErrorHandlingError(
                        "The error handling code threw an exception.",
                        HttpContext.Current.Server.GetLastError());

                    return;
                }
            }

            try
            {
                if (ConfigurationManager.AppSettings.Include(ManualErrorsRenderingMode.ViewResult))
                {
                    // ***** The View Result Approach *****

                    HttpContext.Current.Response.ClearHeaders();
                    HttpContext.Current.Response.ClearContent();
                    HttpContext.Current.Response.StatusCode = statusCode;

                    if (ConfigurationManager.AppSettings.Include(TrySkipIisCustomErrorsMode.Skip))
                    {
                        HttpContext.Current.Response.TrySkipIisCustomErrors = true;
                    }

                    var routeData = new RouteData();

                    routeData.Values.Add("controller", ErrorControllerRouteName);

                    var controllerContext = new ControllerContext(
                        new HttpContextWrapper(HttpContext.Current),
                        routeData, 
                        new FakeController());

                    var viewResult = new ViewResult {ViewName = "ManualErrors"};

                    viewResult.ViewBag.ExceptionMessage = exception.Message;
                    viewResult.ViewBag.StatusCode = HttpContext.Current.Response.StatusCode;
                    viewResult.ViewBag.SessionData = MvcApplication.GetSessionData();

                    viewResult.ExecuteResult(controllerContext);

                    HttpContext.Current.Server.ClearError();
                }
                else if (ConfigurationManager.AppSettings.Include(ManualErrorsRenderingMode.ControllerExecute))
                {
                    // ***** The Controller Execute Approach *****

                    var routeData = new RouteData();

                    routeData.Values.Add("controller", ErrorControllerRouteName);
                    routeData.Values.Add("action", "ManualErrors");
                    routeData.Values.Add("statusCode", statusCode);

                    var requestContext = new RequestContext(new HttpContextWrapper(HttpContext.Current), routeData);

                    var controllerFactory = ControllerBuilder.Current.GetControllerFactory();

                    var controller = controllerFactory.CreateController(requestContext, ErrorControllerRouteName);

                    controller.Execute(requestContext);

                    HttpContext.Current.Server.ClearError();
                }
                else if (ConfigurationManager.AppSettings.Include(ManualErrorsRenderingMode.TransferRequest))
                {
                    // ***** The TransferRequest Approach *****

                    // This method of directing back to MVC does not preserve TempData.

                    if (HttpContext.Current.Session != null)
                    {
                        HttpContext.Current.Session["exception"] = exception;
                    }

                    var url = "~/Error/ManualErrors?statusCode=" + statusCode;

                    HttpContext.Current.Server.TransferRequest(url, false, "GET", null);
                }
                else if (ConfigurationManager.AppSettings.Include(ManualErrorsRenderingMode.MvcHttpHandler))
                {
                    // ***** The MvcHttpHandler Approach *****

                    var path = HttpContext.Current.Request.Path;

                    HttpContext.Current.RewritePath("~/Error/ManualErrors");

                    if (HttpContext.Current.Session != null)
                    {
                        HttpContext.Current.Session["statusCode"] = statusCode;
                    }

                    IHttpHandler handler = new MvcHttpHandler();

                    handler.ProcessRequest(HttpContext.Current);

                    HttpContext.Current.RewritePath(path);

                    HttpContext.Current.Server.ClearError();
                }
            }
            catch (Exception renderException)
            {
                if (ConfigurationManager.AppSettings.Include(ErrorPageExceptionsMode.HandleInApplicationError))
                {
                    HandleErrorHandlingError("The MVC manual error handling code threw an exception.", renderException);
                }
                else
                {
                    throw;
                }
            }
        }

        private static void HandleErrorHandlingError(string reason, Exception exception = null)
        {
            // In this situation, you would most likely log the error then display a static html error page.
            // I'm just writing directly to the response so that I can show the exception details.

            var response = HttpContext.Current.Response;

            response.ClearHeaders();
            response.ClearContent();
            response.StatusCode = (int)HttpStatusCode.InternalServerError;
            HttpContext.Current.Server.ClearError();
            response.Write("YOWZAH!! An unhandled error occurred within Application_Error: " + reason ?? string.Empty);
            response.Write(Environment.NewLine);

            if (exception != null)
            {
                response.Write(exception.Message);
                response.Write(exception.StackTrace);
            }

            HttpContext.Current.ApplicationInstance.CompleteRequest();
        }

        public void Dispose()
        {
        }

        private class FakeController : Controller {}
    }
    // ReSharper restore InconsistentNaming
}