﻿using System;
using System.Configuration;
using System.Net;
using System.Web;
using System.Web.Mvc;
using System.Web.Routing;
using MvcErrorHandling.Controllers;
using MvcErrorHandling.DomainModel.AppSettingsModel;

namespace MvcErrorHandling.MvcHelpers
{
    /// <remarks>
    /// This class closely follows the approach in the ControllerFactoryWrapper class in the NotFoundMvc project.
    /// See https://github.com/andrewdavey/NotFoundMvc.
    /// </remarks>
    public class NotFoundHandlerControllerFactory : DefaultControllerFactory
    {
        protected override IController GetControllerInstance(RequestContext requestContext, Type controllerType)
        {
            // Perform the standard behaviour if we are not manually handling errors using this class.
            if (!ConfigurationManager.AppSettings.Include(ManualNotFoundErrorHandlingMode.WithinMvcFramework))
            {
                return base.GetControllerInstance(requestContext, controllerType);
            }

            try
            {
                var controller = base.GetControllerInstance(requestContext, controllerType);

                var concreteController = controller as Controller;

                if (concreteController != null)
                {
                    concreteController.ActionInvoker = new ActionInvokerWrapper(concreteController.ActionInvoker);
                }

                return controller;
            }
            catch (HttpException exception)
            {
                if (exception.GetHttpCode() == (int)HttpStatusCode.NotFound)
                {
                    // Set up and return the error controller for a 404 error.
                    return CreateNotFoundController(requestContext.RouteData);
                }

                // Allow the exception to rise up the stack if it was not a 404 error.
                throw;
            }
        }

        private static IController CreateNotFoundController(RouteData routeData)
        {
            // Alter the key routing values of the routedata instance passed in.
            routeData.Values["controller"] = "Error";
            routeData.Values["action"] = "ManualErrors";
            routeData.Values["statusCode"] = 404;

            // Note: This controller instance could be created using an IoC container.
            return new ErrorController();
        }

        public static void CreateAndInvokeNotFoundController(HttpContextBase httpContext)
        {
            var routeData = new RouteData();

            var errorController = CreateNotFoundController(routeData);

            errorController.Execute(new RequestContext(httpContext, routeData));
        }

        private class ActionInvokerWrapper : IActionInvoker
        {
            private readonly IActionInvoker _actionInvoker;

            public ActionInvokerWrapper(IActionInvoker actionInvoker)
            {
                if (actionInvoker == null) throw new ArgumentNullException("actionInvoker");

                _actionInvoker = actionInvoker;
            }

            public bool InvokeAction(ControllerContext controllerContext, string actionName)
            {
                try
                {
                    if (!_actionInvoker.InvokeAction(controllerContext, actionName))
                    {
                        CreateAndInvokeNotFoundController(controllerContext.HttpContext);
                    }

                    return true;
                }
                catch (HttpException exception)
                {
                    if (exception.GetHttpCode() == (int)HttpStatusCode.NotFound)
                    {
                        CreateAndInvokeNotFoundController(controllerContext.HttpContext);

                        return true;
                    }

                    throw;
                }
            }
        }
    }
}